package fit;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Vector;

public class StepFixture extends Fixture {
    protected Parse cells;

    // Cross-table state //////////////////////////////////

    // Fixture creates a new StepFixture for each table. We
    // call that evanescent object the "controller". It controls
    // a persistent object (the original controller) that
    // maintains state from table to table. That's the "actor".
    //
    // It's kind of a
    // kludge to have one class do the two things, but it
    // makes the tables somewhat nicer. I think. See
    // ActionFixture for a separation of the two classes


    private StepFixture controller = this;
    private static StepFixture actor;
    private Invocation invocation;  // representation of a row

    protected StepFixture actor() {
        return StepFixture.actor;
    }

    protected void ensureCrossTableStateExists() {
            if (actor() == null) {
                restart();
            }
        }

    // A special form.
    public void restart() {
        StepFixture.actor = this;
    }


    // Traversal //////////////////////////////////

    public void doCells(Parse cells) {
        ensureCrossTableStateExists();
        try {
            if (isSpecialForm(cells)) {
                invocation = new SpecialFormInvocation(controller, cells);
          } else {
                invocation = new EvaluatingInvocation(actor(), cells);
            }
            invocation.execute();
        } catch (Exception e) {
            exception(invocation.cells(), e);
        }
    }

    // Note that each type of fixture could have its own special forms.
    protected boolean isSpecialForm(Parse cells) {
        return cells.text().equals("check") || cells.text().equals("restart");
    }


    // Check is a special form.
    // Check rows have this format:
    // check retriever-method arg0 ... argN expected-value
    public void check() throws NoSuchMethodException {
        Method retrieverMethod = findRetrieverMethod();
        TypeAdapter adapter = TypeAdapter.on(actor(), retrieverMethod.getReturnType());
        check(invocation.lastArgCell(), adapter);
    }

    private Method findRetrieverMethod() throws NoSuchMethodException {
        String retrieverMethodName = invocation.argCellAt(0).text();
        int argCount = invocation.numberOfGivenArgs() - 2;
        return invocation.findMethodByArity(this, retrieverMethodName, argCount);
    }

    public Object valueFromFixture(TypeAdapter a) throws Exception {
        Method retrieverMethod = findRetrieverMethod();
        Object[] args = invocation.argumentValuesInRange(retrieverMethod, 1,
                                                         invocation.numberOfGivenArgs()-1);
        return retrieverMethod.invoke(actor(), args);
    }

}
